﻿using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Collections.Generic;
using System;

namespace HeuristicLab.Problems.MetaOptimization {
  /// <summary>
  /// TODO An operator for analyzing the best solution of Traveling Salesman Problems given in path representation using city coordinates.
  /// </summary>
  [Item("BestParameterConfigurationAnalyzer", "TODO An operator for analyzing the best solution of Traveling Salesman Problems given in path representation using city coordinates.")]
  [StorableClass]
  public sealed class BestParameterConfigurationAnalyzer : SingleSuccessorOperator, IAnalyzer {

    public ScopeTreeLookupParameter<ParameterConfigurationTree> ParameterConfigurationParameter {
      get { return (ScopeTreeLookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public LookupParameter<IRun> BestSolutionParameter {
      get { return (LookupParameter<IRun>)Parameters["BestSolution"]; }
    }
    public ValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (ValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public LookupParameter<DoubleValue> BestKnownQualityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["BestKnownQuality"]; }
    }
    public LookupParameter<IRun> BestKnownSolutionParameter {
      get { return (LookupParameter<IRun>)Parameters["BestKnownSolution"]; }
    }
    public LookupParameter<RunCollection> PopulationParameter {
      get { return (LookupParameter<RunCollection>)Parameters["Population"]; }
    }

    public BestParameterConfigurationAnalyzer() : base() {
      Parameters.Add(new ScopeTreeLookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", "TODO The TSP solutions given in path representation from which the best solution should be analyzed."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", "TODO The qualities of the TSP solutions which should be analyzed."));
      Parameters.Add(new LookupParameter<IRun>("BestSolution", "TODO The best TSP solution."));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", "TODO The result collection where the best TSP solution should be stored."));
      Parameters.Add(new LookupParameter<DoubleValue>("BestKnownQuality", "TODO The quality of the best known solution of this TSP instance."));
      Parameters.Add(new LookupParameter<IRun>("BestKnownSolution", "TODO The best known solution of this TSP instance."));
      Parameters.Add(new LookupParameter<RunCollection>("Population", "TODO The best known solution of this TSP instance."));
    }

    [StorableConstructor]
    private BestParameterConfigurationAnalyzer(bool deserializing) : base(deserializing) { }
    private BestParameterConfigurationAnalyzer(BestParameterConfigurationAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new BestParameterConfigurationAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      ItemArray<DoubleValue> qualities = QualityParameter.ActualValue;
      ResultCollection results = ResultsParameter.ActualValue;
      DoubleValue bestKnownQuality = BestKnownQualityParameter.ActualValue;
      ItemArray<ParameterConfigurationTree> parameterTrees = ParameterConfigurationParameter.ActualValue;

      int i = qualities.Select((x, index) => new { index, x.Value }).OrderBy(x => x.Value).First().index;

      EngineAlgorithm bestAlg = ((EngineAlgorithm)((ParameterConfigurationTree)parameterTrees[i]).ActualValue.Value);
      Run bestRun = new Run(bestAlg);
      ((ParameterConfigurationTree)parameterTrees[i]).CollectResultValues(bestRun.Results);

      if (bestKnownQuality == null || qualities[i].Value < bestKnownQuality.Value) { // todo: respect Maximization:true/false
        BestKnownQualityParameter.ActualValue = new DoubleValue(qualities[i].Value);
        BestKnownSolutionParameter.ActualValue = bestRun;
      }

      IRun best = BestSolutionParameter.ActualValue;
      if (best == null) {
        BestSolutionParameter.ActualValue = bestRun;
        results.Add(new Result("Best Parameter Settings", bestRun));
      } else {
        BestSolutionParameter.ActualValue = bestRun;
        results["Best Parameter Settings"].Value = bestRun;
      }

      // population
      RunCollection rc = new RunCollection();
      foreach (ParameterConfigurationTree pt in parameterTrees.OrderByDescending(x => x.BestQuality.Value)) { // todo: respect Maximization:true/false
        IAlgorithm alg = (IAlgorithm)pt.ActualValue.Value;
        alg.StoreAlgorithmInEachRun = false;
        IRun run = new Run(alg);
        pt.CollectResultValues(run.Results);
        rc.Add(run);
      }
      if (PopulationParameter.ActualValue == null) {
        PopulationParameter.ActualValue = rc;
        results.Add(new Result("Population", rc));
      } else {
        PopulationParameter.ActualValue = rc;
        results["Population"].Value = rc;
      }

      return base.Apply();
    }
  }
}
